# $Id: playfield.py 58 2005-09-03 21:43:40Z  $

### Imports ###
import pygame
import random
from pygame					import sprite
from playfield_capacitors	import *
from playfield_resistor		import PlayfieldResistor
from playfield_led			import PlayfieldLed
from playfield_player		import PlayfieldPlayer

### Constants ###
PLAYFIELD_SPEED = 200
PLAYFIELD_BACKGROUND = "../media/images/pf-background.jpg"
PLAYFIELD_MOVE_KEYS	 = {
	pygame.constants.K_UP		: ( 0, -1),
	pygame.constants.K_DOWN		: ( 0,  1),
	pygame.constants.K_LEFT		: (-1,  0),
	pygame.constants.K_RIGHT	: ( 1,  0),
}
INITIAL_RESISTOR_COUNT	= 10
INITIAL_LED_COUNT		= 5

### Class Playfield ###
class Playfield:
	def __init__(self, playfieldSize, itemFoundCallback):
		"""
		Playfield(itemFoundCallback)
		
		playfieldSize - a (width, height) tuple specifying the size of the PF.
		itemFoundCallback - Will be called when the player hits an item.
		"""

		self.itemFoundCallback	= itemFoundCallback
		self.surface			= pygame.surface.Surface(playfieldSize)
		self.backgroundImage	= pygame.image.load(PLAYFIELD_BACKGROUND)
		self.player				= PlayfieldPlayer()
		self.visibleItems		= sprite.OrderedUpdates()
		self.blockingItems		= sprite.Group()
		self.consumerItems		= sprite.Group()
		self.collectableItems	= sprite.Group()
		self.pressedKeys		= set()
		self.itemClasses		= (
			# Classname					Chance of appearing per second
			(PlayfieldCapacitorBig,		0.5),
			(PlayfieldCapacitorSmall,	1.0),
		)
				
		# Create a few resistors
		for i in xrange(0, INITIAL_RESISTOR_COUNT):
			resistor = PlayfieldResistor()
			self.positionItem(resistor)
			self.blockingItems.add(resistor)
		
		# Put a few leds
		for i in xrange(0, INITIAL_LED_COUNT):
			led = PlayfieldLed()
			self.positionItem(led)
			self.consumerItems.add(led)
		
		# Add the player to visible items group
		self.visibleItems.add(self.player)
		
		# Draw the background image on the surface
		self.surface.blit(self.backgroundImage,
						  self.backgroundImage.get_rect())
	
	def getPlayer(self):
		"""
		getPlayer() -> PlayfieldPlayer
		"""
	
		return self.player

	def getSurface(self):
		"""
		getSurface() -> Surface
		"""
		
		return self.surface
	    
	def handleEvent(self, event):
		"""
		handleEvent(event) -> Bool
		
		Returns True/False whether the event was processed by the playfield.
		"""
		
		if event.type == pygame.KEYDOWN:
			if event.key in PLAYFIELD_MOVE_KEYS:
				self.pressedKeys.add(event.key)
				return True

		if event.type == pygame.KEYUP:
			if event.key in self.pressedKeys:
				self.pressedKeys.remove(event.key)
				return True
		
		return False
		
	def processPressedKeys(self, timePassed):
		"""
		processPressedKeys() -> None
		
		Processes any pressed key.
		"""
		
		for key in self.pressedKeys:
			deltaX, deltaY	= PLAYFIELD_MOVE_KEYS[key]
			deltaX			*= PLAYFIELD_SPEED * timePassed
			deltaY			*= PLAYFIELD_SPEED * timePassed
			oldPosition		= self.player.savePosition()
			self.player.updatePosition(deltaX, deltaY)
			
			# Validate that the player stays within the playfield boundaries
			if not self.surface.get_rect().contains(self.player.rect) :
				self.player.restorePosition(oldPosition)
			
			# Validate the the player does not collide with a blocking item
			if sprite.spritecollideany(self.player, self.blockingItems):
				self.player.restorePosition(oldPosition)
				
			# For each of the consumer items, check if the player collides with it
			collidingConsumers = sprite.spritecollide(self.player, 
													  self.consumerItems,
													  False)
			for consumerItem in self.consumerItems:
				if consumerItem in collidingConsumers:
					consumerItem.updateState(True)
				else:
					consumerItem.updateState(False)
			
			# Check if the player has collected an item
			collectedItems = sprite.spritecollide(self.player,
												  self.collectableItems,
												  True)
			if collectedItems:
				for item in collectedItems:
					self.itemFoundCallback(item)
	
	def doFrame(self, timePassed):
		"""
		doFrame(timePassed) -> None

		Moves the player, according to the amount of time passed
		May call capacicatorCallback(capaci_power)
		The surface should be updated here.
		"""

		self.processPressedKeys(timePassed)
		
		# Update all visible items
		self.visibleItems.update(timePassed)
		
		# Create new items by chance
		for itemClass, itemChance in self.itemClasses:
			if random.random() < itemChance * timePassed:
				pfi = itemClass()
				self.collectableItems.add(pfi)
				self.positionItem(pfi)

		# Update the surface with the changes
		self.visibleItems.clear(self.surface, self.backgroundImage)
		self.visibleItems.draw(self.surface)

	def positionItem(self, item):
		"""
		positionItem(item)
		
		Position an item (of type PlayfieldItem) on the Playfield, making sure
		it doesn't collide with any existing item.
		Also adds the item to the visibleItems group.
		"""
		
		max_x , max_y = self.surface.get_rect().bottomright
		
		max_x -= item.rect.width
		max_y -= item.rect.height

		# Initial positioning
		x, y	= random.uniform(0, max_x), random.uniform(0, max_y)
		item.move(x, y)

		# Make sure the item doesn't collide with any existing visible items
		while sprite.spritecollideany(item, self.visibleItems):
			x, y = random.uniform(0, max_x), random.uniform(0, max_y)
			item.move(x, y)
			
		self.visibleItems.add(item)
